#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h> /* pid_t */
#include <sys/wait.h> 
#include <unistd.h> /* fork() */
#include <regex.h>

#include <config.h>
#include "parse.h"
#include "tree.h"
#include "options.h"
#include "readtags.h"
#include "error.h"

/* open the supplied filename and read it until given line and column number
 * returning the malloc'ed buffer
 * Don't forget to free the string */
static char* read_source_file (const char* filename, int line, int col)
{
	char *buffer = NULL; 
	int file_line = 1;
	int file_col = 1;
		
	FILE *file = fopen (opt_filename, "r");
	if (!file)
	{
		fprintf(stderr, "Error opening file `%s'\n", opt_filename);
		return NULL;
	}

	int c;
	int pos = 0;
	while (1)
	{
		/* increase the buffer size every 1024 bytes */
		if (pos % 1024 == 0)
		{
			buffer = realloc (buffer, ((pos / 1024) + 1) * 1024);
			if (!buffer)
				return NULL;
		}

		/* we found the correct line and column */
		if (file_line == line && file_col == col)
		{
			buffer[pos] = '\0';
			break;
		}

		/* if not keep reading the file */
		c = fgetc(file);
		if (c == EOF)
		{
			free (buffer);
			buffer = NULL;
			break;
		}

		if (c == '\n')
		{
			file_line++;
			file_col = 1;
		}
		else
			file_col++;

		buffer[pos] = c;
		pos++;
	}

	fclose (file);
	return buffer;
}

/* creates a simple hash for all #include lines of a line */
static long calculate_hash(const char* buf)
{
	long res = 0;

	/* find include'ed files and add them to the cache */
	regex_t re;
	regmatch_t pm[2]; // one sub expression -> 2 matches
	memset (&pm, -1, sizeof(pm));
	const char* pattern = "#\\s*include\\s*(<[^>]+>|\"[^\"]+\")";
	char include[256];

	(void) regcomp (&re, pattern, REG_EXTENDED) ;
	while (regexec (&re, buf, 2, &pm[0], 0) == 0) /* while matches found */
	{   
		/* subString found between pm.rm_so and pm.rm_eo */
		int len = pm[1].rm_eo - pm[1].rm_so;
		if (len > 256) len = 256;
		strncpy (include, buf + pm[1].rm_so + 1, len-2); // we don't copy < or ", that's why -1
		include[len-2] = '\0';
		buf += pm[0].rm_eo;

		for (int i=0; i < strlen(include); i++)
			res += (i+1) * include[i];
	}

	/* free allocated memory */
	regfree(&re);

	return res;
}

/* forks and executes ctags to rebuild a tags file
 * storing cache_value in the tags file */
static void build_tags_file(long cache_value)
{
	pid_t child_pid, wait_pid;
	int status = 0;
	switch (child_pid = fork())
	{
		case -1:
			bailout("error fork()'ing child process");
		case 0: /* child */ 
			{
				// create an .icomplete_taglist file to feed the exuberant-ctags program with filenames
				Tree *includes = build_include_tree(NULL, opt_filename);
				FILE *taglist = fopen(".icomplete_taglist", "w");
				if (includes && taglist)
				{
					write_tree_to_file(includes, taglist);
					fclose (taglist);

					/* save the cache information in the tags file */
					FILE *tags = fopen("tags", "w");
					if (tags)
					{
						fprintf(tags, "!_TAG_FILE_FORMAT	2	/extended format; --format=1 will not append ;\" to lines/\n"
								"!_TAG_FILE_SORTED	1	/0=unsorted, 1=sorted, 2=foldcase/\n"
								"!_TAG_PROGRAM_AUTHOR	Darren Hiebert	/dhiebert@users.sourceforge.net/"
								"!_TAG_PROGRAM_NAME	Exuberant Ctags	//\n"
								"!_TAG_PROGRAM_URL	http://ctags.sourceforge.net	/official site/\n"
								"!_TAG_PROGRAM_VERSION	5.5.4	//\n"
								"!_TAG_CACHE\t%ld\n", cache_value);
						fclose(tags);
					}

					char **arguments = (char**)malloc(sizeof(char*) * (8 + (config.cpp_macros.count*2)));
					if (arguments == NULL)
						bailout ("Could not get memory in build_tags_file() for char **arguments");

					arguments[0]=CTAGS_CMD;
					arguments[1]="--append";
					arguments[2]="--language-force=c++";
					arguments[3]="--fields=afiKmsSzn";
					arguments[4]="--c++-kinds=cdefgmnpstuvx";
					arguments[5]="-L";
					arguments[6]=".icomplete_taglist";
					int i = 0;
					for (; i < config.cpp_macros.count; i+=2) 
					{
						arguments[i+7]="-I";
						arguments[i+8]=config.cpp_macros.entries[i];
					}
					arguments[i+7]=NULL;

#if DEBUG >= 2
					printf("CALLING: ");
					for (int x = 0; x <= i+7; x++)
						printf("%s ", arguments[x]);
					printf("\n");
#endif
					// call exuberant-ctags safely by fork() and exec()'ing instead of a simple system()
					if (execvp(CTAGS_CMD, arguments) == -1)
					{
						perror(CTAGS_CMD);
						unlink(".icomplete_taglist");
						bailout("Make sure that exuberant-ctags is correctly installed.\n\nSometimes the executable name is not `ctags', then you need to rebuild the `icomplete' package like this:\n\n# CTAGS_CMD=exuberant-ctags ./configure\n# make\n# sudo make install");
					}
					free_tree(includes);
					printf("%d\n", unlink(".icomplete_taglist"));
				}
				else
					bailout ("Error creating .icomplete_taglist file. Check permissions");
				break;
			}
		default: /* father */
			wait_pid = wait(&status);
			break;
	}
}

int main(int argc, char** argv)
{
	int ret = 0;
	get_cmdine_opts(argc, argv);
	read_config();

	// we have a file name as an argument, parse it
	if (opt_filename)
	{
		// if we supply a filename, line and column numbers are mandatory
		if (opt_line < 1 || opt_column < 1)
			usage();

		char *buf = read_source_file(opt_filename, opt_line, opt_column);
		if (!buf)
			bailout ("Could not read file until given line and column numbers.");

		long cache_value = calculate_hash(buf);
		/* automatic cache mode */
		/* we save a hash value for all included files from the current file
		 * in the tags file, if it matches the hash of the current buffer, nothing
		 * has changed, and reuse the existing tags file */
		if (opt_cache == 1)
		{
			FILE *fCache = fopen("tags", "r");
			bool build_cache = true;
			if (fCache)
			{
				char cacheline[512];
				int maxlines = 10; /* for speed reasons, only search the first 10 lines */
				while (maxlines-- && fgets(cacheline, 512, fCache) != NULL)
				{
					char cmp_pattern[512];
					snprintf(cmp_pattern, 512, "!_TAG_CACHE\t%ld\n", cache_value);
					if (!strcmp(cacheline, cmp_pattern))
					{
#if DEBUG >= 2
						fprintf(stderr, "valid tags file found, icomplete will reuse it\n");
#endif
						build_cache = false;
					}
				}
				fclose(fCache);
			}
			if (build_cache)
				build_tags_file(cache_value);
		}
		/* no cache, always rebuild tags file */
		else if (opt_cache ==2)
		{
			build_tags_file(cache_value);
		}

		/* real parsing starts here */
		Expression exp;
		Scope sc;
		exp = parse_expression(buf, &sc);
		if (exp.access == ParseError)
			bailout("could not parse current expression.");

		free (buf);
		buf = NULL;

#if DEBUG >= 2
		if (exp.access & AccessMembers || exp.access & AccessPointer || exp.access & AccessStatic)
			fprintf(stderr, "class:%s | scope:%s\n", exp.class, sc.scope);
		else if (exp.access & AccessInFunction)
			fprintf(stderr, "class:%s | function:%s | scope:%s\n", strcmp(exp.class, "") ? exp.class : "<none>", exp.function, sc.scope);
		else if (exp.access & AccessGlobal)
			fprintf(stderr, "global functions/variables for scope: %s\n", sc.scope);
#endif

		/* we have all relevant information, so just list the entries */
		if (exp.access == ParseError)
			bailout("Parse error");

		find_entries(&exp, &sc);
	}

	else if (opt_listmembers)
	{
		Scope sc;
		sc.scope[0] = '\0';
		sc.localdef[0] = '\0';
		Expression exp;
		exp.access = AccessMembers;
		strncpy (exp.class, opt_listmembers, 256);
		exp.function[0] = '\0';

		find_entries (&exp, &sc);
	}
	
	free_config();
	return ret;
}

