#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <regex.h>
#include <assert.h>

#include <unistd.h> /* close() */
#include <fcntl.h> /* O_RDONLY */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h> 

#include "tree.h"
#include "parse.h"
#include "options.h"
#include "readtags.h"
#include "error.h"

//brc: new subroutine
static void add_tree_child (Tree* tree, const char* class)
{
/* check this class if not already in the tree to
 * avoid a hangup with a broken (=cirular) tags file */
	if (find_in_tree (tree, class) == NOT_FOUND)
	{
		Tree *child = build_inheritance_tree(class);
		// allocate memory for this element
		tree->inherits = realloc (tree->inherits, (tree->size + 1) * sizeof (Tree));
		tree->inherits[tree->size] = child;
		tree->size++;
	}
}

Tree* build_inheritance_tree (const char* class)
{
	Tree *root = malloc (sizeof(Tree));
	root->name = strdup (class);
	root->size = 0;
    root->inherits = NULL;
	
	tagEntry entry;
	tagFileInfo info;
	tagFile *tfile = tagsOpen ("tags", &info);

//brc: split namespace part and classname
	char nbuffer[256];
	char* nspace=NULL;
	strncpy(nbuffer,class,256);
	nbuffer[255]='\0';
	char* cname=nbuffer;
	char* p=nbuffer;
	char* np=NULL;

 	while ((np=strstr(p,"::"))) 
		p=np+2;
	if (p!=nbuffer) {
		*(p-2)='\0'; 
		nspace=nbuffer;
		cname=p;
	}

	Tree *tree = root;

//brc: namespace workaround: if no namespace is present, try to find the class
//                           also in std::
	if (!nspace) {
		char buffer[512];
		strcpy(buffer,"std::");
		strcat(buffer,cname);
		add_tree_child(tree,buffer);
	}

	int i = 0; // counters
	if (tfile && info.status.opened)
	{
		if (tagsFind (tfile, &entry, cname, TAG_OBSERVECASE | TAG_FULLMATCH) == TagSuccess)
		{
			do
			{
				//brc:follow typedefs ...
				if (!strcmp (entry.kind,"typedef") && entry.address.pattern )
				{
					const char *field_namespace = tagsField(&entry, "namespace"); 
					if ((nspace && field_namespace && !strcmp(field_namespace,nspace))||!nspace)
					{
						char pattern[256];
						strncpy(pattern,entry.address.pattern,256); pattern[255]='\0';
						char* p=strstr(pattern,"typedef");
						if (p && strstr(pattern,cname)) {
							p+=strlen("typedef");
							char* p1;
							if ((p1=strstr(p,"typename"))) p=p1+strlen("typename");
							if (strstr(p,"struct")) continue;
							if ((p=strtok(p," \n\t<"))) {
								free_tree(tree);
								if (nspace) {
									char buffer[512];
									strcpy(buffer,nspace);
									strcat(buffer,"::");
									strcat(buffer,p);
									return build_inheritance_tree(buffer);
								} else {
									return build_inheritance_tree(p);
								}
							}
						}
					}
				}
				/* need to get the entry for the class definition */
				if (!strcmp (entry.kind, "class") 
						|| !strcmp (entry.kind, "struct") 
						|| !strcmp (entry.kind, "union"))
				{
					/* and look for the entry "inherits" to see which are superclasses of class */
					for (; i < entry.fields.count; i++)
					{
						if (!strcmp(entry.fields.list[i].key, "inherits"))
						{
							/* split the inherits string to single classes for multiple inheritance */
							char *allinherits = strdup(entry.fields.list[i].value);
							char *buffer = NULL;
							char *inherit = strtok_r(allinherits, ",", &buffer);
							while (inherit != NULL)
							{
								add_tree_child(tree,inherit);

								/* next token */
								inherit = strtok_r (NULL, ",", &buffer);
							}
							free (allinherits);
							goto end;
						}
					}
					// no more superclasses
					goto end;
				}
			} while (tagsFindNext(tfile, &entry));
		}
end:
		tagsClose (tfile);
	}
	return root;
}

Tree* build_include_tree (Tree *start, const char* filename)
{
	static Tree *root = NULL;
	/* get the full name for the filename */
	if (strlen(filename) < 1)
		return NULL;

	char fullname[FILENAME_MAX];
	int fd = get_filedescriptor(filename, fullname);
	if (fd < 0)
	{
		fprintf (stderr, "TREE: %s NOT found\n", filename);
		close (fd);
		return NULL;
	}

	/* and store it in a Tree* pointer */
	Tree *tree, *child;
	struct stat fileStat;
	char *buffer;

	if (!start)
	{
		tree = malloc (sizeof(Tree));
		tree->name = strdup (fullname);
		tree->size = 0;
		tree->inherits = NULL;
	}
	else
		tree = start;
	if (!root)
		root = tree;


	/* get the file size and memory map the file
	 * make sure to close the file desctiptor because of a maximum
	 * of concurrently allowed open files
	 * MAP_PRIVATE means that write operations only occur in memory
	 */
	if (fstat(fd,&fileStat) < 0)
	{
		perror("Fstat Error");
		return NULL;
	}

	int len = fileStat.st_size, offset = 0;
	void *map = mmap(0, len, PROT_READ, MAP_PRIVATE, fd, offset); 
	close (fd);
	if (map == MAP_FAILED)
	{
		perror ("memmap failed");
		return NULL;
	}
	buffer = map;

	/* find include'ed files and add them to tree */
	regex_t re;
	regmatch_t pm[2]; // one sub expression -> 2 matches
	memset (&pm, -1, sizeof(pm));
	const char* pattern = "#\\s*include\\s*(<[^>]+>|\"[^\"]+\")";
	char include[128];

	(void) regcomp (&re, pattern, REG_EXTENDED) ;
	while (regexec (&re, buffer, 2, &pm[0], 0) == 0) /* while matches found */
	{   
		/* subString found between pm.rm_so and pm.rm_eo */
		int len = pm[1].rm_eo - pm[1].rm_so;
		if (len > 128) len = 128;
		strncpy (include, buffer + pm[1].rm_so + 1, len-2); // we don't copy < or ", that's why -1
		include[len-2] = '\0';
		buffer += pm[0].rm_eo;

		/* get/save full filename for this #include */
		int temp_fd = get_filedescriptor(include, fullname);
		close (temp_fd);
		if (temp_fd < 0 )
			continue;

		/* only insert this include if not already present to
		 * avoid cirular inclusion */
		if (find_in_tree (root, fullname) == NOT_FOUND)
		{
			child = malloc (sizeof(Tree));
			child->name = strdup(fullname);
			child->size = 0;
			child->inherits = 0;
			// allocate memory for this element
			tree->inherits = realloc (tree->inherits, (tree->size + 1) * sizeof (Tree));
			tree->inherits[tree->size] = child;
			tree->size++;

			build_include_tree(child, include);
		}
	}

	/* free allocated memory */
	regfree(&re);
	munmap (map, len);

	return tree;
}

void free_tree(Tree *tree)
{
	int i;
	if (tree)
	{
#if DEBUG >= 2
		fprintf (stderr, "freeing tree: %s\n", tree->name);
#endif
		for (i = 0; i < tree->size; i++)
		{
			free_tree(tree->inherits[i]);
		}
		free (tree->name);
		free (tree->inherits);
		free (tree);
	}
}

void write_tree_to_file(Tree *tree, FILE* file)
{
	fprintf (file, "%s\n", tree->name);
	int i;
	for (i = 0; i < tree->size; i++)
	{
		write_tree_to_file(tree->inherits[i], file);
	}
}

Match find_in_tree(const Tree *tree, const char* class)
{
	int i;
	if (tree)
	{
		if (!strcmp(tree->name, class))
			return DIRECT_HIT;
		for (i = 0; i < tree->size; i++)
		{
			if (find_in_tree(tree->inherits[i], class) >= DIRECT_HIT)
				return SUBCLASS;
		}
	}
	return NOT_FOUND;
}

bool is_member_of_scope (const tagEntry *entry, const Scope *scope, const Tree *tree)
{
	const char *field_class = tagsField(entry, "class");
	const char *field_struct = tagsField(entry, "struct");
	const char *field_union = tagsField(entry, "union");
	//const char *field_namespace = tagsField(entry, "namespace"); 
	const char *field_access = tagsField(entry, "access");

	const char *tag_class = NULL;
	if (field_class)
		tag_class = field_class;
	else if (field_struct)
		tag_class = field_struct;
	else if (field_union)
		tag_class = field_union;
	else
		return false;

	/* tag must be part of (inherited) class to show it */
	Match match_tag = find_in_tree (tree, tag_class);
	if (match_tag == NOT_FOUND)
		return false;
 	if (field_union)
		return true;	
	if (!field_access)
		return false;
	if(!strcmp(field_access, "public"))
		return true;

	/* if access is not public, current scope must be in one of the
	 * (inherited) classes */
	Match match_curscope = find_in_tree (tree, scope->scope);
	if (!strcmp(field_access, "protected") && match_curscope >= DIRECT_HIT)
		return true;
	else if (!strcmp(field_access, "private") && match_curscope == DIRECT_HIT)
		return true;

	return false;
}

void find_entries (const Expression *exp, const Scope *scope)
{
	tagEntry entry;
	tagFileInfo info;
	tagFile *tfile = tagsOpen ("tags", &info);
	char lasttag[256] = ""; /* store last tag to avoid duplicates */

	/* write tags to file instead of stdout */
	FILE *output = stdout;
	if (opt_output)
	{
		output = fopen(opt_output, "w");
		if (!output)
		{
			perror (opt_output);
			return;
		}
	}

	/* create an inheritance tree of our class or current scope, depending what we want to complete */
	Tree *tree = NULL;
	if (exp->access & AccessMembers || exp->access & AccessPointer || exp->access & AccessStatic || exp->access & AccessInFunction)
		tree = build_inheritance_tree(exp->class);
	else if (exp->access & AccessGlobal)
		tree = build_inheritance_tree(scope->scope);

	if (!tree)
		bailout ("Couldn't build inheritance tree\n");


	/* parse the tags file */
	if (tfile && info.status.opened)
	{
		/* we can do a binary search for function definitions */
		if (exp->access == AccessInFunction && tagsFind (tfile, &entry, exp->function, TAG_FULLMATCH | TAG_OBSERVECASE) == TagSuccess)
		{
			do
			{
				bool show_this_tag = false;
				const char *field_kind = tagsField(&entry, "kind");

				if (field_kind && (!strcmp(field_kind, "member") || !strcmp(field_kind, "method")))
					show_this_tag = is_member_of_scope (&entry, scope, tree);
				else if (!strcmp(field_kind, "function") || !strcmp(field_kind, "prototype"))
				{
					if (!strcmp(exp->class, "")) // only a function, not method of a class
						show_this_tag = true;
					else
						show_this_tag = is_member_of_scope (&entry, scope, tree);
				}
				if (!show_this_tag)
					continue;

				/* output the tag if it is valid and no duplicate either to stdout
				 * or to a file if the -o flag was given */
				const char *tag = NULL;
				tag = prettify_tag(&entry, strlen(entry.name) + 5);
				if (tag)
					fprintf(output, "%s\n", tag);
			}
			while ( tagsFindNext (tfile, &entry) == TagSuccess);
		}

		// otherwise loop through all tags
		// and filter those which match one of our inherited class names
		else if (tagsFirst (tfile, &entry) == TagSuccess)
		{
			do
			{
				bool show_this_tag = false;
				const char *field_kind = tagsField(&entry, "kind");
				const char *field_class = tagsField(&entry, "class");
				const char *field_struct = tagsField(&entry, "struct");
				const char *field_union = tagsField(&entry, "union");
				const char *field_namespace = tagsField(&entry, "namespace"); 
				const char *field_access = tagsField(&entry, "access");

				switch (exp->access)
				{
				/*
				 * find MEMBERS of classes/structs
				 */
				case AccessMembers:
				case AccessPointer:
					if (strstr(entry.address.pattern, "static "))
						continue;
					/* Qt 3 defines this, and checking for it does not harm other users */
					if (strstr(entry.address.pattern, "QT_STATIC_CONST "))
						continue;
						
					// don't show destructors, can't call them directly anyways.
					if (entry.name[0] == '~')
					{
						show_this_tag = false;
						break;
					}

					// TODO: namespace support
					if (field_kind && (!strcmp(field_kind, "member") || !strcmp(field_kind, "function") ||
								!strcmp(field_kind, "prototype") || !strcmp(field_kind, "method")))
						show_this_tag = is_member_of_scope (&entry, scope, tree);
					break;
				/*
				 * find STATIC functions/variables (everything after a ::)
				 */
				case AccessStatic:
					/* when we find a tag, the following conditions must be met, to show the tag:
					 * - must be part of a class or namespace
					 * - must contain "static" on the line
					 * - must be "public"
					 */
					if (field_class)
					{
						if (!strcmp (exp->class, field_class) &&
								field_access && !strcmp(field_access, "public"))
							/* XXX: Workaround ctags limitation and try to find out
							 * ourselves, if this tag is a static one, if it 
							 * has "static" in the pattern */
							if (strstr(entry.address.pattern, "static ") ||
									strstr(entry.address.pattern, "QT_STATIC_CONST "))
								show_this_tag = true;
					}
					else if (field_namespace)
					{
						if (!strcmp (exp->class, field_namespace))
							show_this_tag = true;
					}
					break;

				/*
				 * find GLOBAL functions/variables/defines
				 */
				case AccessGlobal:
					/* everthing which is not part of a class or struct matches */
					if (!field_class && !field_struct && !field_union)
					{
						show_this_tag = true;
					}
					else /* also show locals of current scope */
					{
						show_this_tag = is_member_of_scope (&entry, scope, tree);
					}
					break;

				default:
					/* things like if() while() for(), etc. */
					break;
				}

				/* output the tag if it is valid and no duplicate either to stdout
				 * or to a file if the -o flag was given */
				if (show_this_tag && (exp->access == AccessInFunction || strcmp(entry.name, lasttag)))
				{
					const char *tag = NULL;
					if (exp->access & AccessInFunction)
						tag = prettify_tag(&entry, strlen(entry.name) + 5);
					else
					{
						strncpy (lasttag, entry.name, 256);
						tag = prettify_tag(&entry, -1);
					}
					
					if (tag)
						fprintf(output, "%s\n", tag);
				}
			}
			while ( tagsNext (tfile, &entry) == TagSuccess);
		}
		tagsClose(tfile);
	}
	else
	{
		bailout("Could not open `tags' file. You should be able to build a valid one by running this command:\n\nicomplete -l 1 -c 1 <filename>\n\n<filename> should be a source file which contains all #include files you want to have in the tags file");
	}

	free_tree (tree);
	if (output)
		fclose(output);
}
